//--------------------------------------------------------------------------------------------------
//
// full_duplex_serial_vl.c
//
// A full duplex serial driver.
//
// Uses the GCC labels as values extention.
// Runs in a Propeller COG as native code.
//
// Copyright (c) 2011 Michael Rychlik.
// MIT Licensed (see at end of file for exact terms)
//
// History:
//
// 2011-10-08    v0.1  Initial version.
//
//--------------------------------------------------------------------------------------------------

#include "full_duplex_serial_vl.h"

// For Cog regs (_OUTA, _INA, _DIRA..)
#include "cog.h"

// Pointers to mail box in HUB
static _COGMEM volatile struct fds_mailbox* pm;
static _COGMEM volatile int* pTxHead;
static _COGMEM volatile int* pTxTail;
static _COGMEM volatile char* txBuffer;
static _COGMEM volatile int* pRxHead;
static _COGMEM volatile int* pRxTail;
static _COGMEM volatile char* rxBuffer;

// Serial port config
static _COGMEM int txPin;
static _COGMEM int rxPin;
static _COGMEM int bitcycles;
static _COGMEM int halfBitcycles;

// tx_Thread variables
static _COGMEM int txWaitCycles;
static _COGMEM int txShiftReg;
static _COGMEM int txBitCnt;

// rx_thread variables
static _COGMEM int rxWaitCycles;
static _COGMEM int rxShiftReg;
static _COGMEM int rxBitCnt;
static _COGMEM int bit;
static _COGMEM int temp;


static _COGMEM volatile void* pTxThread;
static _COGMEM volatile void* pRxThread;

// Use to define a thread start location
// Saves having ugly labels in the thread code
#define THREAD(name) name :


#define YIELD(line, thisThread, thatThread) \
    thisThread = && next_##line; goto *thatThread; next_##line :;


// Yield execution from the currently running thread to another
// while a given conditin is true.
// Uses the GCC labels as values extention. The task switch here should
// compile to a single Propeller JMPRET instruction when compiled as
// COG native code and optimized.
// Params:
//        line       A unique number for each call.
//        thisThread A void* pointer into the currently running thread.
//        thatThread A void* pointer into the task to swap thread.
//        cond       The condition.
#define WAIT_WHILE(line, thisThread, thatThread, cond)               \
    while(cond)                                                            \
    {                                                             \
        thisThread = && next_##line; goto *thatThread; next_##line :; \
    }        

// Yield execution from the currently running task to another task
// until a given conditin is true.
// Params:
//        line     A unique number for each call.
//        thisTask A void* pointer into the currently running task.
//        thatTask A void* pointer into the task to swap to.
//        cond     The condition.
#define WAIT_UNTIL(line, thisThread, thatThread, cond)               \
    WAIT_WHILE (line, thisThread, thatThread, !(cond))

#define TIME_AFTER(t)  ((int)_CNT - t >= 0)


_NATIVE void main (volatile struct fds_mailbox *m)
{
    // Copy mail box pointer for other functions use.
    pm = m;
    pRxHead = &pm->rxHead;
    pRxTail = &pm->rxTail;
    rxBuffer = pm->rxBuffer;
    pTxHead = &pm->txHead;
    pTxTail = &pm->txTail;
    txBuffer = pm->txBuffer;

    // Set up serial port
    rxPin = pm->rxPin;
    txPin = pm->txPin;
    bitcycles = 80000000 / pm->baudRate;  //<HEATER>Shouldn't CLKFREQ be defined somewhere visible here?</HEATER>
    halfBitcycles = bitcycles / 2;
    _OUTA = (unsigned)1 << txPin;
    _DIRA = (unsigned)1 << txPin;

    // Set up coroutine pointers
    pRxThread = &&rxThread;
    pTxThread = &&txThread;

    // The UART transmitter coroutine.
    THREAD(txThread)
    {
        // We loop forever here.
        while(1)
        {
            // Wait while the TX FIFO is empty
            WAIT_UNTIL(1, pTxThread, pRxThread, *pTxHead != *pTxTail); 

            // Load the tx shift register with a byte from tx FIFO.

            //txShiftReg = txBuffer[(*pTxTail)++ & BUFFER_MASK]; 
            txShiftReg = txBuffer[pm->txTail++ & BUFFER_MASK]; 

            YIELD (11, pTxThread, pRxThread);
            // Output the start bit
            _OUTA = 0;
        
	    // Get the next bit time
            txWaitCycles = (int)_CNT + bitcycles;

            YIELD (12, pTxThread, pRxThread);
            for (txBitCnt = 0; txBitCnt < 8; txBitCnt++)
            {
                WAIT_UNTIL (2, pTxThread, pRxThread, TIME_AFTER(txWaitCycles));  // Wait for the next bit time
                _OUTA = (txShiftReg & 1) << txPin;                         // Output a bit 
                txWaitCycles += bitcycles;                                 // Get next bit time
                YIELD (21, pTxThread, pRxThread);
                txShiftReg >>= 1;                                          // Get next bit into position
                YIELD (22, pTxThread, pRxThread);
            }

            WAIT_UNTIL (3, pTxThread, pRxThread, TIME_AFTER(txWaitCycles));      // Wait for the next bit time
            _OUTA = 1 << txPin;                                            // Output the stop bit 

            // Wait for end  of stop  bit time 
            txWaitCycles += bitcycles;
            WAIT_UNTIL(4, pTxThread,  pRxThread, TIME_AFTER(txWaitCycles));
        }
    }

    // The UART reciever coroutine.
    THREAD(rxThread)
    {
        // We loop forever here.
        while(1)
        {
            rxShiftReg = 0;
            // Wait until a start bit is detected
            WAIT_WHILE (5, pRxThread, pTxThread, ((_INA >> rxPin) & 1));

            // Wait until halfway through start bit
            rxWaitCycles = (int)_CNT + halfBitcycles - 350;
            WAIT_UNTIL (6, pRxThread, pTxThread, TIME_AFTER(rxWaitCycles));

            rxWaitCycles = (int)_CNT + bitcycles - 40;// 100;                           // Get next bit time
            rxBitCnt = 0;        
            do
            { 
                WAIT_UNTIL (7, pRxThread, pTxThread, TIME_AFTER(rxWaitCycles));  // Wait until middle of bit time
                bit = (_INA >> rxPin) & 1;                                  // Read the bit 
                rxWaitCycles = (int)_CNT + bitcycles - 20;                         // Get next bit time               
                rxShiftReg |= bit << rxBitCnt;                              // and place in shift register
            } while (++rxBitCnt < 8);

            YIELD (81, pRxThread, pTxThread);
            // Write the recieved char to the RX FIFO
            if ((*pRxHead - *pRxTail) < BUFFER_SIZE - 1)
            {
            	YIELD (82, pRxThread, pTxThread);
                // Not full: Put char in buffer.
                rxBuffer[(*pRxHead)++ & BUFFER_MASK] = (char)rxShiftReg;
            }
            //rxWaitCycles -= 100;

            // Wait into center of Stop bit 
            WAIT_UNTIL (8, pRxThread, pTxThread, TIME_AFTER(rxWaitCycles));
        }
    }
}
//--------------------------------------------------------------------------------------------------
// TERMS OF USE: MIT License
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//--------------------------------------------------------------------------------------------------

